import uuid
from datetime import timedelta
from unittest.mock import patch

import pytest

from sym.flow.cli.commands.tokens.issue import to_timedelta
from sym.flow.cli.errors import InvalidExpiryError
from sym.flow.cli.symflow import symflow as click_command


class TestTokensIssue:
    @pytest.mark.parametrize(
        "expiry, expected_timedelta",
        [
            ("3s", timedelta(seconds=3)),
            ("30m", timedelta(minutes=30)),
            ("10d", timedelta(days=10)),
            ("1mo", timedelta(days=30)),
        ],
    )
    def test_to_timedelta(self, expiry, expected_timedelta):
        timedelta_int = int(expected_timedelta.total_seconds())
        assert to_timedelta(expiry) == timedelta_int

    @pytest.mark.parametrize(
        "expiry",
        [
            "0s",
            "03s",
            "not a number",
            "3y",
            "m",
        ],
    )
    def test_to_timedelta_invalid(self, expiry):
        with pytest.raises(InvalidExpiryError):
            to_timedelta(expiry)

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_token",
        return_value=str(uuid.uuid4()),
    )
    def test_issue_token(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["tokens", "issue", "--username", "bot-user", "--expiry", "3d"],
            )
            assert result.exit_code == 0

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_token",
        return_value=str(uuid.uuid4()),
    )
    def test_create_token_with_label(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "tokens",
                    "issue",
                    "--username",
                    "bot-user",
                    "--expiry",
                    "3d",
                    "--label",
                    "a label",
                ],
            )
            assert result.exit_code == 0

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_token",
        return_value=str(uuid.uuid4()),
    )
    def test_issue_token_short_options(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["tokens", "issue", "-u", "bot-user", "-e", "3d", "-l", "a label"],
            )
            assert result.exit_code == 0

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_token",
        return_value=str(uuid.uuid4()),
    )
    def test_issue_token_no_newline(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["tokens", "issue", "-u", "bot-user", "-e", "3d", "-l", "a label", "-n"],
            )
            assert "\n" not in result.output

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.create_token",
        return_value="jwt_token",
    )
    def test_issue_token_prompts(self, mock_api, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command, ["tokens", "issue"], input="bot-user\n3d\n"
            )
            assert result.output == "Bot username: bot-user\nExpiry: 3d\njwt_token\n"



from typing import List

import click
import inquirer
from inquirer.questions import Question


def ask(questions: List[Question]):
    answers = inquirer.prompt(questions)
    if not answers:
        raise click.Abort()
    return answers



import copy
from typing import List

import pytest

from sym.flow.cli.helpers.api_operations import OperationType
from sym.flow.cli.helpers.users import UserUpdateSet
from sym.flow.cli.models.user import User
from sym.flow.cli.tests.factories.users import UserFactory
from sym.flow.cli.tests.helpers.sandbox import Sandbox


class TestUserUpdateSet:
    @pytest.fixture
    def user_data(self) -> List[User]:
        users: List[User] = UserFactory.create_batch(5)
        return users

    @pytest.fixture
    def user_update_set(self, user_data) -> UserUpdateSet:
        # NOTE: all users should have the same service slug + external id and therefore the same
        # service keys, so it's fine to just use the first user's services
        yield UserUpdateSet(
            user_data=user_data, service_data=[i.service for i in user_data[0].identities]
        )

    @pytest.fixture
    def edited_user_data(self, user_data):
        edited_data = copy.deepcopy(user_data)
        # Edit user identity
        edited_data[0].identities[1].matcher = {"email": "new_email@symops.io"}
        # Delete user identity
        del edited_data[1].identities[1]
        # Delete an user
        del edited_data[-1]
        # Add an user
        edited_data.append(UserFactory.create())

        return edited_data

    def test_from_dict_init(self, user_data):
        user_set = UserUpdateSet(user_data=user_data)
        assert len(user_set.users) == len(user_data)

    def test_compare_sets(self, user_update_set, edited_user_data):
        edited_set = UserUpdateSet(user_data=edited_user_data)
        result = UserUpdateSet.compare_user_sets(user_update_set, edited_set)

        assert len(result.update_user_ops) == 2
        assert result.update_user_ops[0].operation_type == OperationType.update_user

        assert len(result.delete_identities_ops) == 1
        assert (
            result.delete_identities_ops[0].operation_type
            == OperationType.delete_identity
        )

        assert len(result.delete_user_ops) == 1
        assert result.delete_user_ops[0].operation_type == OperationType.delete_user

    def test_tabulate(self, user_update_set: UserUpdateSet):
        expected_user_tabluation = [
            [identity.to_csv() for identity in u.identities]
            for u in user_update_set.users
        ]
        assert user_update_set.tabulate() == expected_user_tabluation

    def test_write_to_csv(self, user_update_set: UserUpdateSet, sandbox: Sandbox):
        expected_tabulation = [
            [identity.to_csv() for identity in u.identities] + [u.id]
            for u in user_update_set.users
        ]
        headers = ",".join(user_update_set.headers)
        # e.g. [["userid", "id1", "id2"], ["userid", "id3", "id4"]] -> userid,id1,id2\nuserid,id3,id4
        expected_csv_data = "\n".join(
            [",".join(user_data) for user_data in expected_tabulation]
        )
        expected_csv_data = f"{headers}\n{expected_csv_data}\n"

        path = f"{sandbox.path}/foo.csv"
        user_update_set.write_to_csv(path)
        with open(path) as f:
            assert f.read() == expected_csv_data



import click
from inquirer.shortcuts import confirm, text
from policy_sentry.querying.arns import (
    get_matching_raw_arns,
    get_resource_type_name_with_raw_arn,
)
from policy_sentry.util.arns import get_service_from_arn

from sym.flow.cli.commands.policies.helpers import prompt_for_access_levels
from sym.flow.cli.helpers.arn import RawARN
from sym.flow.cli.helpers.iam_policies import IAMPolicies


def prompt_for_arn(policies: IAMPolicies):
    arn = text(message="Enter an ARN", validate=lambda _, x: RawARN.validate(x))
    service = get_service_from_arn(arn)
    resource = get_resource_type_name_with_raw_arn(get_matching_raw_arns(arn)[0])
    click.secho(f"\nFound a {service} {resource}!\n", fg="green")

    levels = prompt_for_access_levels(service)
    policies.add_arn(arn, levels)


@click.command(
    name="generate",
    short_help="Create an AWS IAM Policy from ARNs",
)
def policies_generate():
    policies = IAMPolicies()

    prompt_for_arn(policies)
    while confirm("Would you like to add another ARN?"):
        prompt_for_arn(policies)

    policy = policies.admin_policy()
    click.echo_via_pager(policy)



import json
import threading
import webbrowser
from unittest.mock import patch

import pkce
import pytest
import requests
from requests import Response
from sym.shared.cli.errors import CliError

from sym.flow.cli.errors import InvalidTokenError
from sym.flow.cli.helpers.login.login_flow import BrowserRedirectFlow, LoginError


@pytest.fixture
def browser_redirect_flow(mocker):
    flow = BrowserRedirectFlow(11001)
    mocker.patch.object(pkce, "generate_code_verifier", return_value="test-verifier")
    mocker.patch.object(pkce, "get_code_challenge", return_value="test-challenge")
    mocker.patch.object(webbrowser, "open")
    return flow


@pytest.fixture
def wait_for_code_tester(browser_redirect_flow, global_options):
    def fixture(expected_state: str, expected_code: str, do_test):
        def run_test():
            code = browser_redirect_flow.wait_for_code(expected_state, global_options)
            assert code == expected_code

        th = threading.Thread(target=run_test)
        th.start()
        do_test()
        th.join()

    return fixture


def test_redirect_uri(browser_redirect_flow):
    assert browser_redirect_flow.redirect_url == "http://localhost:11001/callback"


def test_get_auth_code(mocker, browser_redirect_flow, global_options, test_org):
    mocker.patch.object(browser_redirect_flow, "wait_for_code", return_value="1234XYZ")
    mocker.patch.object(browser_redirect_flow, "gen_state", return_value="foo")
    code, verifier = browser_redirect_flow.get_auth_code(global_options, test_org)
    assert code == "1234XYZ"
    assert verifier == "test-verifier"
    webbrowser.open.assert_called_once_with(
        f"https://auth.com/authorize?response_type=code&client_id={test_org.client_id}&code_challenge_method=S256&code_challenge=test-challenge&redirect_uri=http://localhost:11001/callback&scope=admin offline_access&audience=https://api.symops.com&state=foo&prompt=login"
    )


@pytest.mark.skip(reason="flaky, causes circleci to hang sometimes")
def test_wait_for_code(wait_for_code_tester):
    def do_test():
        r = requests.get(
            "http://localhost:11001/callback?code=test_wait_for_code&state=foo"
        )
        assert r.status_code == 200
        assert "Login Successful!" in r.text

    wait_for_code_tester(
        expected_state="foo", expected_code="test_wait_for_code", do_test=do_test
    )


@pytest.mark.skip(reason="flaky, causes circleci to hang sometimes")
def test_wait_for_code_missing_code(wait_for_code_tester):
    def do_test():
        r = requests.get("http://localhost:11001/callback?state=foo")
        assert r.status_code == 401
        assert "Missing code in query" in r.text

    wait_for_code_tester(expected_state="foo", expected_code=None, do_test=do_test)


@pytest.mark.skip(reason="flaky, causes circleci to hang sometimes")
def test_wait_for_code_missing_state(wait_for_code_tester):
    def do_test():
        r = requests.get("http://localhost:11001/callback?code=test_wait_for_code")
        assert r.status_code == 401
        assert "Missing state in query" in r.text

    wait_for_code_tester(expected_state="", expected_code=None, do_test=do_test)


@pytest.mark.skip(reason="flaky, causes circleci to hang sometimes")
def test_wait_for_code_invalid_state(wait_for_code_tester):
    def do_test():
        r = requests.get(
            "http://localhost:11001/callback?code=test_wait_for_code&state=bar"
        )
        assert r.status_code == 401
        assert "Invalid state" in r.text

    wait_for_code_tester(expected_state="foo", expected_code=None, do_test=do_test)


def test_get_token_from_code_ok(mocker, browser_redirect_flow, global_options, test_org):
    response_json = json.loads(
        """
        {
        "access_token": "test-access-token",
        "token_type": "Bearer",
        "expires_in": "86400",
        "refresh_token": "test-refresh-token",
        "scope": "test-scope offline_access"
    }
    """
    )
    response = Response()
    response.status_code = 200

    mocker.patch.object(requests, "post", return_value=response)
    mocker.patch.object(response, "json", return_value=response_json)
    token = browser_redirect_flow.get_token_from_code(
        global_options, test_org, "1234XYZ", "test-verifier"
    )
    requests.post.assert_called_once_with(
        "https://auth.com/oauth/token",
        headers={"content-type": "application/x-www-form-urlencoded"},
        data={
            "grant_type": "authorization_code",
            "client_id": test_org.client_id,
            "code_verifier": "test-verifier",
            "code": "1234XYZ",
            "redirect_uri": "http://localhost:11001/callback",
        },
    )
    assert token["access_token"] == "test-access-token"
    assert token["refresh_token"] == "test-refresh-token"


def test_get_token_from_code_missing_refresh_token(
    mocker, browser_redirect_flow, global_options, test_org
):
    with pytest.raises(InvalidTokenError):
        response_json = json.loads(
            """
            {
            "access_token": "test-access-token",
            "token_type": "Bearer",
            "expires_in": "86400",
            "scope": "test-scope offline_access"
        }
        """
        )
        response = Response()
        response.status_code = 200
        mocker.patch.object(requests, "post", return_value=response)
        mocker.patch.object(response, "json", return_value=response_json)
        browser_redirect_flow.get_token_from_code(
            global_options, test_org, "1234XYZ", "test-verifier"
        )


def test_get_token_from_code_error_response(
    mocker, browser_redirect_flow, global_options, test_org
):
    with pytest.raises(LoginError):
        response = Response()
        response.status_code = 403
        mocker.patch.object(requests, "post", return_value=response)
        browser_redirect_flow.get_token_from_code(
            global_options, test_org, "1234XYZ", "test-verifier"
        )


def test_login(mocker, global_options, browser_redirect_flow, auth_token, test_org):
    mocker.patch.object(
        browser_redirect_flow,
        "get_auth_code",
        return_value=("test-code", "test-verifier"),
    )
    mocker.patch.object(
        browser_redirect_flow, "get_token_from_code", return_value=auth_token
    )
    token = browser_redirect_flow.login(global_options, test_org)
    assert token == auth_token
    browser_redirect_flow.get_auth_code.assert_called_once_with(global_options, test_org)
    browser_redirect_flow.get_token_from_code.assert_called_once_with(
        global_options, test_org, "test-code", "test-verifier"
    )


@patch(
    "sym.flow.cli.helpers.login.login_flow.BrowserRedirectFlow.get_auth_code",
    return_value=(None, None),
)
def test_login_no_auth_code(
    mock_get_auth_code, browser_redirect_flow, global_options, test_org
):
    with pytest.raises(CliError, match="Unable to get access token from Auth0"):
        browser_redirect_flow.login(global_options, test_org)

    mock_get_auth_code.assert_called_once()



import re
from functools import cached_property
from typing import Dict, List, Optional

from policy_sentry.querying.arns import get_matching_raw_arns


class RawARN:
    _all_defaults = {}

    @classmethod
    def with_shared_defaults(cls, template: str):
        return cls(template, cls._all_defaults)

    @classmethod
    def validate(cls, arn: str):
        try:
            return bool(get_matching_raw_arns(arn))
        except:
            return False

    def __init__(self, template: str, defaults: Optional[Dict[str, str]] = None) -> None:
        self.template = template
        self._defaults = defaults or {}

    @cached_property
    def fields(self) -> List[str]:
        return re.findall(r"\$\{(\w+)\}", self.template)

    @property
    def defaults(self) -> Dict[str, str]:
        return {"Partition": "aws", **self._defaults}

    def set(self, field: str, value: str):
        self._defaults[field] = value

    def update(self, values: Dict[str, str]):
        self._defaults.update(values)

    def values(self) -> Dict[str, str]:
        return {field: self.defaults.get(field, "") for field in self.fields}

    def format(self):
        return self.template.replace("$", "").format(**self.values())

    def __str__(self):
        return self.format()



from typing import Any, Dict, Optional, TypedDict

from sym.shared.cli.helpers.config.base import ConfigBase

from sym.flow.cli.models import AuthToken, Organization


class ConfigSchema(TypedDict, total=False):
    org: str
    client_id: str
    email: str
    auth_token: AuthToken


def deepget(key: str, data: Dict[str, Any]) -> Optional[str]:
    """Tries to get a nested value from a dict"""
    for k in key.split("."):
        try:
            data = data[k]
        except KeyError:
            return None

    return data if isinstance(data, str) else None


class Config(ConfigBase[ConfigSchema]):
    @classmethod
    def get_value(cls, key: str) -> Optional[str]:
        config = cls.instance()
        return deepget(key, config)

    @classmethod
    def get_org(cls) -> Organization:
        config = cls.instance()
        return Organization(slug=config["org"], client_id=config["client_id"])

    @classmethod
    def is_logged_in(cls):
        return "org" in cls.instance()

    @classmethod
    def get_access_token(cls) -> Optional[str]:
        try:
            return cls.instance()["auth_token"]["access_token"]
        except KeyError:
            return None

    @classmethod
    def logout(cls):
        if not cls.is_logged_in():
            return

        cfg = cls.instance()
        del cfg["email"]
        del cfg["org"]
        del cfg["client_id"]
        del cfg["auth_token"]


def store_login_config(email: str, org: Organization, auth_token: AuthToken) -> str:
    cfg = Config.instance()
    cfg["email"] = email
    cfg["org"] = org.slug
    cfg["client_id"] = org.client_id
    cfg["auth_token"] = auth_token
    return str(cfg.file)



from typing import TYPE_CHECKING, Dict, List, Union

import click
from sym.shared.cli.errors import CliError, CliErrorWithHint

if TYPE_CHECKING:
    # Importing here due to circular dependencies
    from sym.flow.cli.models.service_type import ServiceType


class LoginError(CliError):
    def __init__(self, response_body) -> None:
        super().__init__(f"Error logging in: {response_body}")


class UnknownOrgError(CliError):
    def __init__(self, slug: str) -> None:
        super().__init__(f"Unknown organization for slug: {slug}")


class UnknownUserError(CliError):
    def __init__(self, email: str) -> None:
        super().__init__(f"Unknown user for email: {email}")


class UnknownBotError(CliError):
    def __init__(self, username: str) -> None:
        super().__init__(f"Unknown bot: {username}")


class UserAlreadyExists(CliError):
    def __init__(self, email: str) -> None:
        super().__init__(f"A user already exists with email: {email}")


class BotAlreadyExists(CliError):
    def __init__(self, username: str) -> None:
        super().__init__(f"A bot already exists with username: {username}")


class InvalidTokenError(CliError):
    def __init__(self, raw_token) -> None:
        super().__init__(f"Unable to parse token: {raw_token}")


class NotAuthorizedError(CliError):
    def __init__(self) -> None:
        super().__init__(
            "You are not authorized to perform this action. Please run `symflow login`"
        )


class InvalidExternalIdError(CliErrorWithHint):
    def __init__(self, service_type: "ServiceType", external_id: str) -> None:
        error_message = click.style(
            f"The external ID '{external_id}' is invalid for service type '{service_type.type_name}'",
            fg="red",
        )

        super().__init__(
            error_message,
            service_type.help_str,
        )


class MissingServiceError(CliErrorWithHint):
    def __init__(self, service_type: str) -> None:
        error_message = click.style(
            f"No service is registered for type {service_type}",
            fg="red",
        )

        super().__init__(
            error_message,
            f"You can create the service with `symflow services create --service-type {service_type}`",
        )


class MissingIdentityValueError(CliErrorWithHint):
    def __init__(self, identifier: str, command: str = "users") -> None:
        error_message = click.style(
            f"Identity value cannot be empty",
            fg="red",
        )

        super().__init__(
            error_message,
            f"If you want to delete the identity, run `symflow {command} delete-identity {identifier}`",
        )


class InvalidChoiceError(CliErrorWithHint):
    def __init__(self, value: str, valid_choices: List[str]) -> None:
        error_message = click.style(
            f"Invalid input: '{value}'",
            fg="red",
        )

        super().__init__(
            error_message,
            f"Try one of: {', '.join(valid_choices)}",
        )


class MissingChoicesError(CliError):
    def __init__(self) -> None:
        error_message = click.style(
            f"No choices were provided!",
            fg="red",
        )

        super().__init__(error_message)


class ReferencedObjectError(CliError):
    def __init__(self, references: Dict[str, List[str]]) -> None:
        counts = " and ".join(f"{len(refs)} {name}" for name, refs in references.items())

        error_message = click.style(
            f"Cannot perform delete because it is referenced by {counts}",
            fg="red",
        )

        super().__init__(error_message)


class InvalidExpiryError(CliErrorWithHint):
    def __init__(self, expiry: str) -> None:
        error_message = click.style(
            f"Invalid expiry input: {expiry}",
            fg="red",
        )

        super().__init__(
            error_message,
            f"Accepted values are a non-zero integer followed by s, m, d, or mo. e.g. 3d",
        )


class SymAPIRequestError(CliErrorWithHint):
    def __init__(self, message: str, request_id: str) -> None:
        super().__init__(
            "An API error occurred!",
            message
            + click.style(
                f"\n\nPlease contact support and include your Request ID ({request_id}).\nhttps://docs.symops.com/docs/support",
                fg="white",
                bold=True,
            ),
        )


class SymAPIAggregateError(SymAPIRequestError):
    def __init__(self, errors: Union[str, List[str]], request_id: str) -> None:
        self.errors = errors if isinstance(errors, list) else [errors]
        message = "\n\n".join([error for error in self.errors])
        super().__init__(message, request_id)


class SymAPIMissingEntityError(SymAPIRequestError):
    error_codes = [404]

    def __init__(self, response_code: int, request_id: str) -> None:
        super().__init__(f"Missing entity ({response_code}).", request_id)


class SymAPIError(SymAPIRequestError):
    def __init__(self, message: str, code: str, request_id: str) -> None:
        super().__init__(f"An unexpected error occurred ({code}): {message}", request_id)


class SymAPIUnknownError(SymAPIRequestError):
    def __init__(self, response_code: int, request_id: str) -> None:
        super().__init__(
            f"An unknown error with status code {response_code}.", request_id
        )



from typing import Optional

import click

from sym.flow.cli.commands.users.utils import get_or_prompt_service
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.identity_operations import delete_identity


@click.command(
    name="delete-identity",
    short_help="Delete an Identity for a User",
)
@click.argument("email", required=True, type=str)
@click.option(
    "--service-type",
    "_service_type",
    help="The type of service the identity is tied to",
)
@click.option(
    "--external-id",
    "_external_id",
    help="The identifier for the service",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def delete_user_identity(
    options: GlobalOptions,
    email: str,
    _service_type: Optional[str],
    _external_id: Optional[str],
) -> None:
    """For an existing Sym User, delete a single Identity such as an AWS SSO Principal UUID or a Slack User ID.

    \b
    Example:
        `symflow users delete-identity user@symops.io`

    """
    original_user = options.sym_api.get_user(email)
    service = get_or_prompt_service(options.sym_api, _service_type, _external_id)
    delete_identity(options, original_user, service)



import uuid

import click

from sym.flow.cli.errors import BotAlreadyExists
from sym.flow.cli.helpers.api_operations import (
    Operation,
    OperationHelper,
    OperationSets,
    OperationType,
)
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models import BaseService, Identity, ServiceType, User
from sym.flow.cli.models.user_type import UserType


@click.command(name="create", short_help="Create a new Sym Bot User")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("username")
def bots_create(options: GlobalOptions, username: str) -> None:
    """
    Creates a new bot-user with the given username
    """
    bot_users = options.sym_api.get_users({"type": UserType.BOT})
    if username in [u.sym_identifier for u in bot_users]:
        raise BotAlreadyExists(username)

    identities = [
        Identity(
            service=BaseService(slug=ServiceType.SYM.type_name, external_id="cloud"),
            matcher={"username": username},
        )
    ]

    operations = OperationSets(
        # update_user also handles create user
        update_user_ops=[
            Operation(
                operation_type=OperationType.update_user,
                original_value=None,
                new_value=User(
                    id=str(uuid.uuid4()), identities=identities, type=UserType.BOT
                ),
            )
        ],
        delete_identities_ops=[],
        delete_user_ops=[],
    )

    operation_helper = OperationHelper(options, operations)
    options.sym_api.update_users(operation_helper.update_users_payload)
    click.secho(f"Successfully created {username}!", fg="green")



from sym.flow.cli.commands.services.click.inquirer_choice import InquirerChoiceOption
from sym.flow.cli.commands.services.click.service_type_choice import ServiceTypeChoice
from sym.flow.cli.models.service_type import ServiceType


class ServiceTypeOption(InquirerChoiceOption):
    """Custom click.Option that uses inquirer to prompt users for service types.
    Choices can be set with the `service_type_choices` kwarg, otherwise it defaults to all public ServiceTypes.
    """

    def __init__(self, *args, **kwargs):
        kwargs["inquirer_choices"] = kwargs.get("inquirer_choices", ServiceType.public())
        super().__init__(*args, **kwargs)
        self.type = ServiceTypeChoice(choices=kwargs["inquirer_choices"])



from .auth import AuthToken, UserCredentials, parse_auth_token
from .organization import Organization
from .service import BaseService, Service
from .service_type import ServiceType
from .user import CSVMatcher, Identity, User



import click

from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models import ServiceType


@click.command(name="delete", short_help="Delete a Sym Bot User")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("username")
def bots_delete(options: GlobalOptions, username: str) -> None:
    """
    Deletes a Bot User with the given username and revokes all their tokens.
    """
    payload = {
        "users": [
            {
                "identity": {
                    "service_type": ServiceType.SYM.type_name,
                    "matcher": {"username": username},
                },
            }
        ]
    }

    options.sym_api.delete_user(payload)
    click.secho(f"Successfully deleted bot user {username}!", fg="green")



import click

from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="list", short_help="List all domains")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def domains_list(options: GlobalOptions) -> None:
    """List your organization's configured domains.

    Domains are used to decide who can join your organization. For example,
    when an unknown user interacts with your Sym Slack App, they will be automatically
    added to your organization (and allowed to run your Flows) if and only if their email
    domain matches one defined for your organization.

    Users may still be added manually using `symflow users create` even if their email does not
    match any of the configured domains.
    """

    org = options.sym_api.get_current_organization()

    if domains := org.domains:
        click.echo("Your organization's configured domains:")
        click.echo("- " + "\n- ".join(domains))
    else:
        click.echo("Your organization does not have any domains configured.")



from dataclasses import dataclass
from typing import Optional
from urllib.parse import urlparse

from sym.shared.cli.data.global_options_base import GlobalOptionsBase

from sym.flow.cli.helpers.constants import DEFAULT_API_URL, DEFAULT_AUTH_URL


@dataclass
class GlobalOptions(GlobalOptionsBase):
    api_url: str = DEFAULT_API_URL
    auth_url: str = DEFAULT_AUTH_URL
    access_token: Optional[str] = None

    @property
    def sym_api(self):
        from sym.flow.cli.helpers.api import SymAPI

        return SymAPI(self.api_url, self.access_token)

    def set_access_token(self, sym_jwt):
        self.access_token = sym_jwt.strip()

    def set_api_url(self, url):
        url = url.rstrip("/")
        if not self._validate_url(url):
            raise ValueError("Specified API URL '{}' is not valid.".format(url))

        self.api_url = url

    def set_auth_url(self, url):
        url = url.rstrip("/")
        if not self._validate_url(url):
            raise ValueError("Specified Auth URL '{}' is not valid.".format(url))

        self.auth_url = url

    def _validate_url(self, url):
        try:
            parts = urlparse(url)
            if parts.scheme not in ("http", "https"):
                return False
            if not parts.netloc:
                return False
            if parts.fragment or parts.query or parts.params:
                return False
            return True
        except ValueError:
            return False

    def to_dict(self):
        return {
            **super().to_dict(),
            "api_url": self.api_url,
            "auth_url": self.auth_url,
        }



from unittest.mock import patch

from sym.flow.cli.symflow import symflow as click_command


class TestUsersDelete:
    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete(
        self,
        mock_delete_user,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command, ["users", "delete", "sym", "--user-id", "user@symops.io"]
            )
            assert result.exit_code == 0

        mock_delete_user.assert_called_once()

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete_missing_arguments(
        self,
        _,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command, ["users", "delete", "--user-id", "user@symops.io"]
            )
            assert result.exit_code == 2

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete_unmapped_service(
        self,
        mock_delete_user,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "delete", "unknown", "--user-id", '{"id": "user@symops.io"}'],
            )
            assert result.exit_code == 0

        mock_delete_user.assert_called_once()

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete_unmapped_service_invalid_json(
        self,
        _,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "delete", "unknown", "--user-id", "user@symops.io"],
            )
            assert result.exit_code == 1
            assert isinstance(result.exception, ValueError)



"""Config

Allows users to get values from their Sym Flow config
"""

import click
from sym.shared.cli.helpers.sym_group import SymGroup

from sym.flow.cli.helpers.global_options import GlobalOptions

from .config_get import config_get
from .config_set import config_set


@click.group(cls=SymGroup, short_help="Interact with the Sym config file", hidden=True)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def config(options: GlobalOptions) -> None:
    """Operations on the config file"""
    # For internal use only


config.add_command(config_get)
config.add_command(config_set)



from typing import List, Union, cast
from unittest.mock import patch

import pytest
from sym.shared.cli.errors import CliError

from sym.flow.cli.errors import InvalidTokenError, LoginError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.login.login_flow import LoginFlow
from sym.flow.cli.models import AuthToken, Organization
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.conftest import get_mock_response

TEST_LOGIN_ORG_SLUG = "test"


def mock_login_flow(result: Union[AuthToken, CliError]):
    class MockLoginFlow(LoginFlow):
        def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
            # TypedDict doesn't support isinstance
            if isinstance(result, dict):
                return result
            raise cast(CliError, result)

        def gen_browser_login_params(self, options: GlobalOptions, org: Organization):
            return ("", None, None)

    return MockLoginFlow()


@pytest.fixture
def login_flow(auth_token):
    return mock_login_flow(auth_token)


@pytest.fixture
def login_flow_login_error():
    return mock_login_flow(LoginError("test"))


@pytest.fixture
def login_flow_invalid_token_error():
    return mock_login_flow(InvalidTokenError("test"))


login_test_org = Organization(slug="test", client_id="12345abc")


@pytest.fixture
def login_tester(click_setup):
    def tester(
        login_flow: LoginFlow,
        org_slug: str,
        expected_output: List[str],
        success: bool = True,
    ):
        with patch.object(LoginFlow, "get_login_flow", return_value=login_flow):
            with click_setup() as runner:
                result = runner.invoke(click_command, ["login", "--org-slug", org_slug])
                print(result.output)
                for expected_str in expected_output:
                    assert expected_str in result.output
                if success:
                    assert result.exit_code == 0
                else:
                    assert result.exit_code != 0

    return tester


@patch(
    "sym.flow.cli.helpers.api.SymAPI.get_organization_from_slug",
    return_value=login_test_org,
)
def test_login_ok(mock_get_org, login_flow, login_tester):
    login_tester(
        login_flow=login_flow,
        org_slug=TEST_LOGIN_ORG_SLUG,
        success=True,
        expected_output=["Login succeeded"],
    )

    mock_get_org.assert_called_once_with(TEST_LOGIN_ORG_SLUG)


@patch(
    "sym.flow.cli.helpers.api.SymAPI.get_organization_from_slug",
    return_value=login_test_org,
)
def test_login_login_error(_, login_flow_login_error, login_tester):
    login_tester(
        login_flow=login_flow_login_error,
        org_slug=TEST_LOGIN_ORG_SLUG,
        success=False,
        expected_output=[f"Error logging in: {TEST_LOGIN_ORG_SLUG}"],
    )


@patch(
    "sym.flow.cli.helpers.api.SymAPI.get_organization_from_slug",
    return_value=login_test_org,
)
def test_login_invalid_token_error(_, login_flow_invalid_token_error, login_tester):
    login_tester(
        login_flow=login_flow_invalid_token_error,
        org_slug=TEST_LOGIN_ORG_SLUG,
        success=False,
        expected_output=[f"Unable to parse token: {TEST_LOGIN_ORG_SLUG}"],
    )


@patch(
    "requests.request",
    return_value=(
        get_mock_response(
            404,
            {
                "message": f"No Organization with the slug {TEST_LOGIN_ORG_SLUG} was found."
            },
        )
    ),
)
@patch("sym.flow.cli.helpers.rest.SymRESTClient.make_headers", return_value={})
def test_login_unknown_org(_, __, login_flow, login_tester):
    login_tester(
        login_flow=login_flow,
        org_slug=TEST_LOGIN_ORG_SLUG,
        success=False,
        expected_output=[f"No Organization with the slug {TEST_LOGIN_ORG_SLUG}"],
    )



import click

from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="revoke", short_help="Revoke a Sym Token")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("jti")
def tokens_revoke(options: GlobalOptions, jti: str) -> None:
    """
    Revokes a token with the given identifier
    """
    options.sym_api.revoke_token(jti)
    click.secho(
        f"Successfully revoked token with identifier {jti}",
        fg="green",
    )



from datetime import datetime, timezone
from typing import Callable, Dict, List, Optional, TypeVar

import click
import inquirer

from sym.flow.cli.errors import InvalidChoiceError

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")


def filter_dict(
    d: Dict[KeyT, ValueT], filter_func: Callable[[ValueT], bool] = lambda v: v
) -> Dict[KeyT, ValueT]:
    return {k: v for k, v in d.items() if filter_func(v)}


def get_or_prompt(value: Optional[str], prompt: str, choices: List[str]) -> str:
    sorted_choices = sorted(choices)
    if not value and len(choices) == 1:
        click.echo(f"{prompt}: Using '{choices[0]}'")
        return choices[0]
    elif not value:
        return inquirer.list_input(prompt, choices=sorted_choices)
    elif value not in choices:
        raise InvalidChoiceError(value=value, valid_choices=sorted_choices)

    return value


def utc_to_local(utc_dt: datetime):
    return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def human_friendly(dt: datetime) -> str:
    """Returns a human-friendly datetime string. Includes timezone information.
    e.g. '2021-11-19 04:40:35 EST'
    """
    return dt.strftime("%Y-%m-%d %X %Z")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flow',
 'flow.cli',
 'flow.cli.commands',
 'flow.cli.commands.bots',
 'flow.cli.commands.config',
 'flow.cli.commands.debug',
 'flow.cli.commands.domains',
 'flow.cli.commands.policies',
 'flow.cli.commands.resources',
 'flow.cli.commands.services',
 'flow.cli.commands.services.click',
 'flow.cli.commands.services.hooks',
 'flow.cli.commands.tokens',
 'flow.cli.commands.users',
 'flow.cli.helpers',
 'flow.cli.helpers.login',
 'flow.cli.models',
 'flow.cli.tests',
 'flow.cli.tests.commands',
 'flow.cli.tests.commands.bots',
 'flow.cli.tests.commands.config',
 'flow.cli.tests.commands.debug',
 'flow.cli.tests.commands.resources',
 'flow.cli.tests.commands.services',
 'flow.cli.tests.commands.services.hooks',
 'flow.cli.tests.commands.tokens',
 'flow.cli.tests.commands.users',
 'flow.cli.tests.factories',
 'flow.cli.tests.helpers',
 'flow.cli.tests.helpers.iam_policies',
 'flow.cli.tests.helpers.login',
 'flow.cli.tests.models']

package_data = \
{'': ['*']}

install_requires = \
['analytics-python>=1.3.1,<2.0.0',
 'auth0-python>=3.23.1,<4.0.0',
 'boto3>=1.16.20,<2.0.0',
 'click>=8.0.0,<9.0.0',
 'immutables>=0.14,<0.15',
 'inflection>=0.5.1,<0.6.0',
 'inquirer>=2.7.0,<3.0.0',
 'mistune<2.0.0',
 'pkce>=1.0,<2.0',
 'policy-sentry==0.9.0',
 'prompt-toolkit>=3.0.21,<4.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pytest-factoryboy>=2.1.0,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'semver>=2.13.0,<3.0.0',
 'sentry-sdk>=0.19.3,<0.20.0',
 'sym-shared-cli>=0.2.2,<0.3.0',
 'tabulate>=0.8.7,<0.9.0',
 'validators>=0.18.1,<0.19.0',
 'xattr>=0.9.7,<0.10.0']

entry_points = \
{'console_scripts': ['symflow = sym.flow.cli.symflow:symflow']}

setup_kwargs = {
    'name': 'sym-flow-cli',
    'version': '2.0.1',
    'description': "Sym's Official CLI for Implementers",
    'long_description': '# sym-flow-cli\n\nThis is the official CLI for [Sym](https://symops.com/) Implementers. Check out the docs [here](https://docs.symops.com/docs/install-sym-flow).\n',
    'author': 'SymOps, Inc.',
    'author_email': 'pypi@symops.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://symops.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List, Optional

from pydantic import BaseModel


class Organization(BaseModel):
    id: Optional[str] = None
    slug: str
    client_id: Optional[str] = None
    domains: List[str] = []



from typing import Optional

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.utils import get_or_prompt
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType


def get_or_prompt_service(
    api: SymAPI, service_type: Optional[str] = None, external_id: Optional[str] = None
) -> Service:
    """Given optional _service_type and _external_id inputs, return a BaseService that exists in the organization

    If _service_type or _external_id are not provided, prompts the user for input.
    Validates that the provided service type + external ID correspond with a Service in the organization
    """
    # Get only services that are already registered with the organization
    services = api.get_services(service_types=ServiceType.all_public_names())
    existing_service_types = list(set(s.service_type for s in services))

    # Prompt the user for an existing service or validate their input
    service_type = get_or_prompt(
        service_type, "Which service type?", existing_service_types
    )

    # Prompt the user for the external ID of the requested service or validate their input
    external_ids = list(
        set(s.external_id for s in services if s.service_type == service_type)
    )
    external_id = get_or_prompt(
        external_id, f"Which {service_type} service?", external_ids
    )

    return next(
        s
        for s in services
        if s.service_type == service_type and s.external_id == external_id
    )



import click

from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(name="remove", short_help="Remove a domain")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("domain")
def domains_remove(options: GlobalOptions, domain: str) -> None:
    """Remove a domain from your organization.

    Note: this will not affect any existing users in your organization.
    """

    options.sym_api.remove_domain(domain)
    click.echo(f"{domain} successfully removed as a domain.")



import shutil
import tempfile
from io import BytesIO
from pathlib import Path

from sym.shared.cli.decorators import require_bins, run_subprocess

from sym.flow.cli.helpers.global_options import GlobalOptions

from .boto import path_for_uploading, put_public_fileobj


@require_bins("terraform")
@run_subprocess
def terraform_output(directory: str):
    yield ("terraform", f"-chdir={directory}", "output", "-json")


def do_upload_terraform_output(dir: str, options: GlobalOptions) -> str:
    directory = Path(dir).resolve()
    output = terraform_output(
        directory,
        run_subprocess_options_=options,
        capture_output_=True,
    )
    txid, object_path = path_for_uploading(directory / "output", "json")
    put_public_fileobj(BytesIO(output[0].encode()), object_path)
    return txid


def do_upload_directory(dir: str) -> str:
    directory = Path(dir).resolve()
    txid, object_path = path_for_uploading(directory, "zip")
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / txid
        shutil.make_archive(str(path), format="zip", root_dir=str(directory))
        put_public_fileobj(path.with_suffix(".zip").open("rb"), object_path)
        return txid


def do_upload_file(file: str) -> str:
    path = Path(file).resolve()
    txid, object_path = path_for_uploading(path, path.suffix)
    put_public_fileobj(path.open("rb"), object_path)
    return txid



import json
from pathlib import Path

import pytest

from sym.flow.cli.helpers.iam_policies import IAMPolicies

# From the following TF:
#
# resource "sym_resources" "production_bucket" {
#   label = "Prod"
#   arn = "arn:aws:s3:::foobar"
# }
#


@pytest.fixture
def expected_policy_json():
    def _expected_policy_json(name):
        path = Path(__file__).resolve().parent / f"expected_{name}_policy.json"
        return path.read_text()

    return _expected_policy_json


def _prepare_policy(policy):
    return sorted(json.loads(policy)["Statement"], key=lambda x: x["Sid"])


def assert_policies(policy1, policy2):
    assert _prepare_policy(policy1) == _prepare_policy(policy2)


def test_iam_policies(expected_policy_json):
    policies = IAMPolicies(name="SymAccess")
    policies.add_arn("arn:aws:s3:::foobar")

    assert_policies(policies.read_policy(), expected_policy_json("read"))
    assert_policies(policies.write_policy(), expected_policy_json("write"))
    assert_policies(policies.admin_policy(), expected_policy_json("admin"))



SegmentWriteKey = "awgf20oKYlsxq73mW6lO6iVyCOgepPIr"
SentryDSN = "https://6a8722276347467b801e5aa152a48ed1@o373009.ingest.sentry.io/5462801"

DEFAULT_API_URL = "https://api.symops.com/api/v1"
DEFAULT_AUTH_URL = "https://symops.us.auth0.com"
DEFAULT_SCOPES = ["admin", "offline_access"]

WRITE_ONLY_AWS_ACCESS_KEY_ID = "AKIA23WDB56I6QOC4ZAR"
WRITE_ONLY_AWS_SECRET_ACCESS_KEY = "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"

DROPBOX_BUCKET = "sym-dropbox-us-east-1"


CLI_PACKAGE_NAME = "sym-flow-cli"



from unittest.mock import patch

import pytest

from sym.flow.cli.errors import InvalidTokenError
from sym.flow.cli.helpers.login.login_flow import LoginError, PasswordPromptFlow
from sym.flow.cli.models import UserCredentials
from sym.flow.cli.tests.conftest import get_mock_response

PASSWORD_PROMPT_PATH = "sym.flow.cli.helpers.login.login_flow.PasswordPromptFlow"


@pytest.fixture
def password_prompt_flow(mocker):
    return PasswordPromptFlow()


@pytest.fixture
def user_creds():
    return UserCredentials("username", "password")


def test_login_with_user_creds(
    global_options, password_prompt_flow, user_creds, test_org
):
    response = get_mock_response(
        status_code=200,
        data={
            "access_token": "test-access-token",
            "token_type": "Bearer",
            "expires_in": "86400",
            "refresh_token": "test-refresh-token",
            "scope": "test-scope offline_access",
        },
    )

    with patch("requests.post", return_value=response) as mock_post:
        token = password_prompt_flow.login_with_user_creds(
            global_options,
            user_creds,
            test_org,
        )

    mock_post.assert_called_once_with(
        "https://auth.com/oauth/token",
        headers={"content-type": "application/x-www-form-urlencoded"},
        data={
            "grant_type": "http://auth0.com/oauth/grant-type/password-realm",
            "realm": "Username-Password-Authentication",
            "username": "username",
            "password": "password",
            "audience": "https://api.com",
            "scope": "admin offline_access",
            "client_id": test_org.client_id,
        },
    )

    assert token["access_token"] == "test-access-token"
    assert token["refresh_token"] == "test-refresh-token"


def test_login_with_user_creds_no_refresh_token(
    global_options, password_prompt_flow, user_creds, test_org
):
    with pytest.raises(InvalidTokenError):
        response = get_mock_response(
            status_code=200,
            data={
                "access_token": "test-access-token",
                "token_type": "Bearer",
                "expires_in": "86400",
                "scope": "test-scope offline_access",
            },
        )

        with patch("requests.post", return_value=response):
            password_prompt_flow.login_with_user_creds(
                global_options,
                user_creds,
                test_org,
            )


def test_login_with_user_creds_error_response(
    global_options, password_prompt_flow, user_creds, test_org
):
    with pytest.raises(LoginError):
        with patch("requests.post", return_value=get_mock_response(status_code=403)):
            password_prompt_flow.login_with_user_creds(
                global_options,
                user_creds,
                test_org,
            )


def test_login(global_options, password_prompt_flow, user_creds, auth_token, test_org):
    with patch(
        f"{PASSWORD_PROMPT_PATH}.prompt_for_user_credentials", return_value=user_creds
    ) as mock_prompt:
        with patch(
            f"{PASSWORD_PROMPT_PATH}.login_with_user_creds", return_value=auth_token
        ) as mock_login:
            token = password_prompt_flow.login(global_options, test_org)

    assert token == auth_token
    mock_prompt.assert_called_once()
    mock_login.assert_called_once_with(global_options, user_creds, test_org)



import uuid
from unittest.mock import patch

from sym.flow.cli.commands.services.hooks.slack_delete import slack_delete
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.constants import DEFAULT_API_URL


class TestSlackDeleteHook:
    """Suite for testing Slack uninstallation."""

    @patch("sym.flow.cli.helpers.api.SymAPI.uninstall_slack")
    @patch("sym.flow.cli.commands.services.hooks.slack_delete.click.echo")
    def test_click_calls_uninstall_method_and_echoes(
        self, mock_click_echo, mock_slack_uninstall
    ):
        fake_service_id = str(uuid.uuid4())
        slack_delete(SymAPI(DEFAULT_API_URL), fake_service_id)

        mock_slack_uninstall.assert_called_once_with(fake_service_id)
        mock_click_echo.assert_called_once_with(
            "Uninstall successful! The Sym App has been removed from your Slack workspace."
        )



from unittest.mock import call, patch

import pytest

from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import ServiceFactory, UserFactory


@patch("sym.flow.cli.helpers.api_operations.OperationHelper.handle_update_users")
class TestUsersUpdateIdentity:
    @pytest.fixture
    def users(self):
        yield UserFactory.create_batch(3)

    @pytest.fixture
    def services(self, users):
        # Create a second slack service
        slack2 = ServiceFactory.create(slug=ServiceType.SLACK.type_name)
        all_services = [
            i.service
            for i in users[0].identities
            if i.service.service_type != ServiceType.SYM.type_name
        ] + [slack2]

        # User Factory creates the same services for each user
        yield all_services

    @pytest.fixture(autouse=True)
    def patchypatch(self, users, services):
        with patch("sym.flow.cli.helpers.api.SymAPI.get_users", return_value=users):
            with patch(
                "sym.flow.cli.helpers.api.SymAPI.get_services", return_value=services
            ):
                yield

    def test_update_user_identity(self, mock_apply, click_setup, users):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "users",
                    "update-identity",
                    users[0].sym_email,
                    "--new-value",
                    "U123",
                    "--service-type",
                    "slack",
                    "--external-id",
                    "T12345",
                ],
            )
            assert result.exit_code == 0

    @pytest.mark.parametrize(
        "service_type, external_id, invalid_input, hints",
        [
            ("pagerduty", "unknown-id", "unknown-id", "pytest-pd"),  # bad external_id
            (
                "unknown-service",
                "T12345",
                "unknown-service",
                "aptible, aws_iam, aws_sso, github, okta, pagerduty, slack",
            ),  # bad service_type
            (
                "auth0",
                "T12345",
                "auth0",
                "aptible, aws_iam, aws_sso, github, okta, pagerduty, slack",
            ),  # service_type not registered with org
        ],
    )
    def test_update_user_identity_bad_input(
        self,
        mock_apply,
        service_type,
        external_id,
        invalid_input,
        hints,
        click_setup,
        users,
    ):
        with click_setup() as runner:
            # Test unknown external_id
            result = runner.invoke(
                click_command,
                [
                    "users",
                    "update-identity",
                    users[0].sym_email,
                    "--new-value",
                    "U123",
                    "--service-type",
                    service_type,
                    "--external-id",
                    external_id,
                ],
            )
            assert f"Invalid input: '{invalid_input}'" in result.output
            assert f"Try one of: {hints}" in result.output

    def test_update_user_identity_unknown_email(
        self,
        mock_apply,
        click_setup,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "update-identity", "someone@symops.io", "--new-value", "U123"],
            )
            assert "Unknown user for email: someone@symops.io" in result.output

    def test_update_user_identity_prompt_both(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            # Prompts both service type and external ID
            slack_external_ids = sorted(
                [
                    s.external_id
                    for s in services
                    if s.service_type == ServiceType.SLACK.type_name
                ]
            )

            with patch(
                "inquirer.list_input", side_effect=["slack", slack_external_ids[0]]
            ) as mock_inquire:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--new-value",
                        "U123",
                    ],
                )
                assert result.exit_code == 0

                mock_inquire.assert_has_calls(
                    [
                        call(
                            "Which service type?",
                            choices=sorted(list({s.service_type for s in services})),
                        ),
                        call("Which slack service?", choices=slack_external_ids),
                    ]
                )

    def test_update_user_identity_prompt_external_id(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            # Prompts external ID
            slack_external_ids = sorted(
                [
                    s.external_id
                    for s in services
                    if s.service_type == ServiceType.SLACK.type_name
                ]
            )
            with patch(
                "inquirer.list_input", side_effect=[slack_external_ids[0]]
            ) as mock_inquire:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--new-value",
                        "U123",
                        "--service-type",
                        "slack",
                    ],
                )
                assert result.exit_code == 0
                mock_inquire.assert_has_calls(
                    [call("Which slack service?", choices=slack_external_ids)]
                )

    def test_update_user_identity_prompt_service_type(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            # Prompts service type
            with patch("inquirer.list_input", side_effect=["slack"]) as mock_inquire:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--new-value",
                        "U123",
                        "--external-id",
                        "T12345",
                    ],
                )
                assert result.exit_code == 0
                mock_inquire.assert_has_calls(
                    [
                        call(
                            "Which service type?",
                            choices=sorted(list({s.service_type for s in services})),
                        )
                    ]
                )

    def test_update_user_identity_does_not_prompt_external_id(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            # Does not prompt external_id if there is only one service
            with patch("inquirer.list_input", side_effect=["pagerduty"]) as mock_inquire:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--new-value",
                        "U123",
                    ],
                )
                assert result.exit_code == 0
                mock_inquire.assert_has_calls(
                    [
                        call(
                            "Which service type?",
                            choices=sorted(list({s.service_type for s in services})),
                        )
                    ]
                )

                # Should print the auto-selected external ID
                assert "Which pagerduty service?: Using 'pytest-pd'" in result.output

    def test_update_user_identity_prompt_new_value(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            with patch("inquirer.text", return_value="P1234") as mock_inquire_text:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--service-type",
                        "slack",
                        "--external-id",
                        "T12345",
                    ],
                )
                assert result.exit_code == 0

                mock_inquire_text.assert_has_calls([call(message="New value?")])

    def test_update_user_identity_blank_value(
        self, mock_apply, click_setup, users, services
    ):
        with click_setup() as runner:
            with patch("inquirer.text", return_value="") as mock_inquire_text:
                result = runner.invoke(
                    click_command,
                    [
                        "users",
                        "update-identity",
                        users[0].sym_email,
                        "--service-type",
                        "slack",
                        "--external-id",
                        "T12345",
                    ],
                )
                assert result.exit_code != 0
                assert "Identity value cannot be empty" in result.output



from unittest.mock import patch

import pytest

from sym.flow.cli.commands.users.list_identities import get_user_data
from sym.flow.cli.errors import NotAuthorizedError
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.users import UserUpdateSet
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import UserFactory

COMMAND_PATH = "sym.flow.cli.commands.users.list_identities"


class TestUsersListIdentities:
    @pytest.fixture
    def user_update_set(self) -> UserUpdateSet:
        yield UserUpdateSet(user_data=UserFactory.create_batch(2))

    @patch(f"{COMMAND_PATH}.click.echo")
    @patch(f"{COMMAND_PATH}.tabulate_user_data", return_value="some data")
    def test_command_echoes_tabulated_data_without_file_path(
        self, mock_tabulate, mock_click_echo, click_setup, user_update_set
    ):
        with patch(
            f"{COMMAND_PATH}.get_user_data", return_value=user_update_set
        ) as mock_get_user:
            with click_setup() as runner:
                result = runner.invoke(click_command, ["users", "list-identities"])
                assert result.exit_code == 0

        mock_get_user.assert_called_once()
        mock_tabulate.assert_called_once_with(user_update_set)
        mock_click_echo.assert_called_once_with("some data")

    @patch(f"{COMMAND_PATH}.click.echo")
    @patch(f"{COMMAND_PATH}.tabulate_user_data", return_value="some data")
    @patch("sym.flow.cli.helpers.users.UserUpdateSet.write_to_csv")
    def test_command_saves_to_file_with_file_path(
        self,
        mock_write_to_csv,
        mock_tabulate,
        mock_click_echo,
        click_setup,
        user_update_set,
    ):
        with patch(
            f"{COMMAND_PATH}.get_user_data", return_value=user_update_set
        ) as mock_get_user:
            with click_setup() as runner:
                result = runner.invoke(
                    click_command,
                    ["users", "list-identities", "--output-file", "pytest.csv"],
                )
                assert result.exit_code == 0

        mock_get_user.assert_called_once()
        mock_write_to_csv.assert_called_once_with("pytest.csv")
        mock_tabulate.assert_not_called()
        mock_click_echo.assert_called_once_with("Saved 2 users to pytest.csv.")

    @patch(
        f"{COMMAND_PATH}.get_user_data",
        side_effect=ValueError("random error"),
    )
    def test_click_call_catches_unknown_error(
        self, mock_get_integration_data, click_setup
    ):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "list-identities"])
            assert result.exit_code == 1
            assert isinstance(result.exception, ValueError)
            assert str(result.exception) == "random error"

        mock_get_integration_data.assert_called_once()

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.get_users",
        side_effect=NotAuthorizedError,
    )
    def test_users_list_not_authorized_errors(
        self,
        mock_get_users,
    ):
        with pytest.raises(NotAuthorizedError, match="symflow login"):
            get_user_data(SymAPI("http://fake.symops.com/api/v1"))
        mock_get_users.assert_called_once()



import csv
import uuid
from copy import copy
from functools import cached_property
from typing import TYPE_CHECKING, Dict, List, Optional

from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import MANAGED_SERVICES, ServiceType
from sym.flow.cli.models.user import Identity, User

if TYPE_CHECKING:
    from sym.flow.cli.helpers.api_operations import OperationSets


class UserUpdateSet:
    def __init__(
        self, service_data: List[Service] = None, user_data: List[User] = None
    ) -> None:
        self.users: List[User] = user_data or []
        self.services: List[Service] = service_data or []

    @property
    def uneditable_service_types(self) -> List[str]:
        """The list of service types which may at any point be set or edited
        by users.
        """
        service_types = copy(MANAGED_SERVICES)

        # Note: ServiceType.SYM is not technically allowed to be updated, but
        # must be included in the editable list so that new users may be created.
        service_types.remove(ServiceType.SYM.type_name)
        return service_types

    def tabulate(self, include_user_id: bool = False) -> List[List[str]]:
        """Returns contained User data formatted as a list of lists
        where the inner lists contain an item for each of the User's
        Identities.

        include_user_id: Boolean indicating whether or not to include the User ID in the returned row data.
        The User ID is required when writing `symflow users list-identities` to CSV, but hidden when printed
        to the terminal

        e.g. if the Identities are Sym, Slack, Pagerduty:
            [["user@symops.io", "U12345", "P12345"]]
        """
        result = []
        for user in self.users:
            user_data = []
            for service in self.services:
                if service.service_type not in self.uneditable_service_types:
                    identity_repr = user.identity_repr_for_service(service)
                    user_data.append(identity_repr)
            if include_user_id:
                user_data.append(user.id)
            result.append(user_data)
        return result

    def write_to_csv(self, file_path: str) -> None:
        """Writes User data to a CSV file at the path provided."""
        with open(file_path, "w") as f:
            writer = csv.writer(f)
            writer.writerow(self.headers)
            for user_data in self.tabulate(include_user_id=True):
                writer.writerow(user_data)

    def add_csv_row(self, row: Dict) -> None:
        user_id = row.pop("User ID", "")
        # If no user id is empty, assume a temporary uuid for user creation
        if user_id == "":
            user_id = str(uuid.uuid4())
        user = User(id=user_id, identities=[])

        for service_key, matcher_value in row.items():
            if service_key and matcher_value:
                user.identities.append(
                    Identity.from_csv(
                        matcher_value=matcher_value, service_key=service_key
                    )
                )

        self.users.append(user)

    @classmethod
    def compare_user_sets(
        cls, original: "UserUpdateSet", edited: "UserUpdateSet"
    ) -> "OperationSets":
        from sym.flow.cli.helpers.api_operations import (
            Operation,
            OperationSets,
            OperationType,
        )

        old_users = {user.id for user in original.users}
        new_users = {user.id for user in edited.users}

        update_user_ops = []
        delete_user_identity_operations = []
        delete_user_ops = []

        created_users = new_users - old_users
        updated_users = new_users.intersection(old_users)
        deleted_users = old_users - new_users

        for user_id in created_users:
            new_user = edited.get_user_by_id(user_id)
            operation = Operation(
                operation_type=OperationType.update_user,
                original_value=new_user,
                new_value=new_user,
            )
            update_user_ops.append(operation)

        for user_id in updated_users:
            old_user = original.get_user_by_id(user_id)
            new_user = edited.get_user_by_id(user_id)

            old_idents = {
                i.service.service_key
                for i in old_user.filtered_identities(MANAGED_SERVICES)
            }
            new_idents = {
                i.service.service_key
                for i in new_user.filtered_identities(MANAGED_SERVICES)
            }

            created_or_updated_idents = new_idents.intersection(old_idents) | (
                new_idents - old_idents
            )
            deleted_idents = old_idents - new_idents

            for service_key in created_or_updated_idents:
                old_identity = old_user.get_identity_from_key(service_key)
                new_identity = new_user.get_identity_from_key(service_key)

                if old_identity == new_identity:
                    continue

                # Either User doesn't have this identity yet or has this identity but the value is different, update the user
                operation = Operation(
                    operation_type=OperationType.update_user,
                    original_value=old_user,
                    new_value=new_user,
                )

                # operation includes all identities for a user, only append it once
                update_user_ops.append(operation)
                break

            if deleted_idents:
                # At least one identity has been deleted from the set, update the user
                operation = Operation(
                    operation_type=OperationType.delete_identity,
                    original_value=old_user,
                    new_value=new_user,
                )

                # operation includes all identities for a user, only append it once
                delete_user_identity_operations.append(operation)

        for user_id in deleted_users:
            user_to_delete = original.get_user_by_id(user_id)

            # User doesn't exist anymore, delete it
            operation = Operation(
                operation_type=OperationType.delete_user, original_value=user_to_delete
            )
            delete_user_ops.append(operation)

        return OperationSets(
            update_user_ops=update_user_ops,
            delete_identities_ops=delete_user_identity_operations,
            delete_user_ops=delete_user_ops,
        )

    def get_user_by_id(self, user_id: str) -> Optional[User]:
        return next((u for u in self.users if u.id == user_id), None)

    @cached_property
    def headers(self) -> List[str]:
        """The list of headers from the CSV file."""
        return [
            s.service_key
            for s in self.services
            if s.service_type not in self.uneditable_service_types
        ] + ["User ID"]



from typing import List
from uuid import UUID

import click
from tabulate import tabulate

from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.utils import human_friendly, utc_to_local
from sym.flow.cli.models import User


@click.command(name="list", short_help="List all Sym Tokens")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def tokens_list(options: GlobalOptions) -> None:
    """
    Lists all Sym Tokens in your Organization generated by `symflow tokens issue`
    """
    table_data = get_tokens_data(options.sym_api)
    click.echo(tabulate(table_data, headers="firstrow"))


def get_tokens_data(api: SymAPI) -> List[List[str]]:
    tokens = api.get_tokens()

    user_cache = {}
    table_data = [["Token ID", "User", "Created By", "Expires At", "Last Used", "Label"]]
    for t in tokens:
        user = get_user(t.user.id, user_cache, api)
        created_by = get_user(t.created_by.id, user_cache, api)
        table_data.append(
            [
                t.identifier,
                user.sym_identifier,
                created_by.sym_identifier,
                human_friendly(utc_to_local(t.expires_at)),
                human_friendly(utc_to_local(t.updated_at)),
                t.label,
            ]
        )

    return table_data


def get_user(user_id: UUID, user_cache: dict, api: SymAPI) -> User:
    if user := user_cache.get(user_id):
        return user
    else:
        user = user_cache[user_id] = api.get_user_by_id(user_id)

    return user



import click

from sym.flow.cli.commands.services.click.external_id_option import ExternalIdOption
from sym.flow.cli.commands.services.click.service_type_option import ServiceTypeOption
from sym.flow.cli.commands.services.hooks.slack_new import slack_new
from sym.flow.cli.errors import InvalidExternalIdError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType


def post_service_create_hooks(options: GlobalOptions, service: Service) -> None:
    """Registered hooks to call after successful creation of services"""

    if service.service_type == ServiceType.SLACK.type_name:
        slack_new(options.sym_api, service.id)


@click.command(name="create", short_help="Set up a new service")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--service-type",
    help="The service to set up",
    prompt=True,
    required=True,
    cls=ServiceTypeOption,
)
@click.option(
    "--external-id",
    help="The identifier for the service",
    prompt=True,
    required=True,
    cls=ExternalIdOption,
)
def services_create(
    options: GlobalOptions, service_type: ServiceType, external_id: str
) -> None:
    """Set up a new service for your organization."""

    # If a validator is defined, validate the external ID.
    if service_type.validator and not service_type.validator(external_id):
        raise InvalidExternalIdError(service_type, external_id)

    service = options.sym_api.create_service(service_type, external_id)

    click.secho(
        f"Successfully set up service type {service_type.type_name} with external ID {external_id}!",
        fg="green",
    )

    post_service_create_hooks(options, service)



from unittest.mock import call, patch

import pytest

from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import ServiceFactory, UserFactory


class TestServicesDelete:
    @pytest.fixture
    def services(self):
        users = UserFactory.create_batch(3)
        # Create a second slack service
        slack2 = ServiceFactory.create(slug=ServiceType.SLACK.type_name)
        all_services = [
            i.service
            for i in users[0].identities
            if i.service.service_type != ServiceType.SYM.type_name
        ] + [slack2]

        # User Factory creates the same services for each user
        yield all_services

    @pytest.fixture(autouse=True)
    def patchypatch(self, services):
        with patch("sym.flow.cli.helpers.api.SymAPI.get_services", return_value=services):
            yield

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.get_service_references",
        return_value={"identities": [], "integrations": []},
    )
    @patch("sym.flow.cli.helpers.api.SymAPI.delete_service")
    @patch("sym.flow.cli.commands.services.services_delete.pre_service_delete_hooks")
    def test_services_delete(
        self,
        mock_pre_delete_hooks,
        mock_delete_service,
        mock_get_service_references,
        click_setup,
        services,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "services",
                    "delete",
                    "--service-type",
                    services[0].service_type,
                    "--external-id",
                    services[0].external_id,
                ],
            )
            assert result.exit_code == 0
            assert (
                f"Successfully deleted service type slack with external ID {services[0].external_id}!"
                in result.output
            )
            mock_pre_delete_hooks.assert_called_once()
            mock_pre_delete_hooks.call_args.args[1] == services[0]
            mock_delete_service.assert_called_once()

    def test_services_delete_prompt_external_id(self, services, click_setup):
        with click_setup() as runner:
            with patch("inquirer.list_input") as mock_inquire:
                result = runner.invoke(
                    click_command,
                    ["services", "delete", "--service-type", "slack"],
                )

            slack_external_ids = sorted(
                [
                    s.external_id
                    for s in services
                    if s.service_type == ServiceType.SLACK.type_name
                ]
            )
            mock_inquire.assert_has_calls(
                [call("Which slack service?", choices=slack_external_ids)]
            )

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_service")
    def test_services_delete_not_found_errors(self, mock_delete_service, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "services",
                    "delete",
                    "--service-type",
                    "slack",
                    "--external-id",
                    "T123ABC",
                ],
            )
            mock_delete_service.assert_not_called()
            assert result.exit_code is not 0
            assert "Invalid input: 'T123ABC" in result.output
            assert "Try one of:" in result.output

    @patch(
        "sym.flow.cli.helpers.api.SymAPI.get_service_references",
        return_value={"identities": ["uuid1"], "integrations": ["uuid2"]},
    )
    @patch("sym.flow.cli.helpers.api.SymAPI.delete_service")
    @patch("sym.flow.cli.commands.services.services_delete.pre_service_delete_hooks")
    def test_services_delete_has_references(
        self,
        mock_pre_delete_hooks,
        mock_delete_service,
        mock_get_service_references,
        click_setup,
        services,
    ):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                [
                    "services",
                    "delete",
                    "--service-type",
                    services[0].service_type,
                    "--external-id",
                    services[0].external_id,
                ],
            )
            mock_get_service_references.assert_called_once()
            mock_pre_delete_hooks.assert_not_called()
            mock_delete_service.assert_not_called()
            assert result.exit_code is not 0
            assert (
                f"Cannot perform delete because it is referenced by 1 identities and 1 integrations"
                in result.output
            )



import factory

from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.user import Identity, User
from sym.flow.cli.models.user_type import UserType


class ServiceFactory(factory.Factory):
    class Meta:
        model = Service

    id = factory.Faker("uuid4")
    label = factory.Sequence(lambda n: "Service %03d" % n)
    slug = factory.Faker("word", ext_word_list=[s.type_name for s in ServiceType])
    external_id = factory.Sequence(lambda n: "ABC-%03d" % n)


class IdentityFactory(factory.Factory):
    class Meta:
        model = Identity

    service = factory.SubFactory(ServiceFactory)
    matcher = factory.Dict({"email": "user@symops.io"})
    profile = factory.Dict(
        {"first_name": "Guinea", "last_name": "Pig", "email": "user@symops.io"}
    )


class UserFactory(factory.Factory):
    class Meta:
        model = User

    id = factory.Faker("uuid4")
    identities = factory.List(
        [
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.SYM.type_name,
                service__external_id="cloud",
                matcher=factory.Dict(
                    {"email": factory.Sequence(lambda n: "user-%03d@symops.io" % n)}
                ),
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.SLACK.type_name,
                service__external_id="T12345",
                matcher={"user_id": "U12345"},
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.PAGERDUTY.type_name,
                service__external_id="pytest-pd",
                matcher={"user_id": "PD123"},
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.AWS_SSO.type_name,
                service__external_id="sso-inst-arn/12345",
                matcher={
                    "principal_uuid": "1111111111-22222222-3333-4444-5555-666666666666"
                },
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.AWS_IAM.type_name,
                service__external_id="1234567890",
                matcher={"user_arn": "arn:aws:iam::838419636750:user/user@symops.io"},
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.APTIBLE.type_name,
                service__external_id="123-456-7890",
                matcher={"user_id": "123"},
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.OKTA.type_name,
                service__external_id="https://domain.okta.org",
                matcher={"user_id": "123"},
            ),
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.GITHUB.type_name,
                service__external_id="github-org-test",
                matcher={"user_id": "123"},
            ),
        ]
    )
    type = UserType.NORMAL
    created_at = factory.Faker("date_time")
    is_admin = False


class BotUserFactory(UserFactory):
    identities = factory.List(
        [
            factory.SubFactory(
                IdentityFactory,
                service__slug=ServiceType.SYM.type_name,
                service__external_id="cloud",
                matcher=factory.Dict(
                    {"username": factory.Sequence(lambda n: "botuser-%03d" % n)}
                ),
            ),
        ]
    )
    type = UserType.BOT
